/**
 * Copyright (C) Mellanox Technologies Ltd. 2015-2016.  ALL RIGHTS RESERVED.
 * See file LICENSE for terms.
 */

#include <coll_test.h>
#include <sys/mman.h>
#include <malloc.h>

#define MAX_NB_REQS (512)
/* sharp supported datatype info*/
struct sharp_test_data_types_t sharp_test_data_types [] =
{
	{"UINT_32_BIT",	SHARP_DTYPE_UNSIGNED,	MPI_UNSIGNED,	MPI_2INT, 4},
	{"INT_32_BIT",	SHARP_DTYPE_INT,	MPI_INT,	MPI_2INT, 4},
	{"UINT_64_BIT",	SHARP_DTYPE_UNSIGNED_LONG, MPI_UNSIGNED_LONG, MPI_LONG_INT, 8},
	{"INT_64_BIT",	SHARP_DTYPE_LONG,	MPI_LONG,	MPI_LONG_INT, 8},
	{"FLOAT_32_BIT",SHARP_DTYPE_FLOAT,	MPI_FLOAT,	MPI_FLOAT_INT, 4},
	{"FLOAT_64_BIT",SHARP_DTYPE_DOUBLE,	MPI_DOUBLE,	MPI_DOUBLE_INT, 8},
	{"NULL", SHARP_DTYPE_NULL, 0, 0, 0}
};

/*sharp supported reduce op info */
struct sharp_op_type_t sharp_test_ops [] =
{
	{"MAX", SHARP_OP_MAX, MPI_MAX},
	{"MIN", SHARP_OP_MIN, MPI_MIN},
	{"SUM", SHARP_OP_SUM, MPI_SUM},
//	{"PROD", SHARP_OP_PROD, MPI_PROD},
	{"LAND", SHARP_OP_LAND, MPI_LAND},
	{"BAND", SHARP_OP_BAND, MPI_BAND},
	{"LOR", SHARP_OP_LOR, MPI_LOR},
	{"BOR",SHARP_OP_BOR, MPI_BOR},
	{"LXOR", SHARP_OP_LXOR, MPI_LXOR},
	{"BXOR", SHARP_OP_BXOR, MPI_BXOR},
	{"MAXLOC", SHARP_OP_MAXLOC, MPI_MAXLOC},
	{"MINLOC", SHARP_OP_MINLOC, MPI_MINLOC},
	{"NOOP", SHARP_OP_NULL, 0}
};

static void test_data_template_init(struct data_template *template, long value)
{
	template->unsigned_int_val = (unsigned int) value;
	template->int_val = (int) value;
	template->unsigned_long_val = (unsigned long) value;
	template->long_val = (long) value;
	template->float_val = (float) value;
	template->double_val = (double) value;
	/*use rank as tag for min_loc/max_loc ops */
	template->int_loc_tag_val = (int)value -1;
	template->int_min_max_loc_result = 0;
}

static void test_init_buffer(enum sharp_datatype data_type, int length,
			     struct data_template *template, void *buffer,
			     enum sharp_reduce_op op)
{
	int i;
	void *temp;

	for (i = 0; i < length; i++) {
		switch(data_type) {
		case SHARP_DTYPE_INT:
			((int *)buffer)[i] = template->int_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				((int *)buffer)[++i] = template->int_loc_tag_val;
			}
			break;
		case SHARP_DTYPE_LONG:
			((long *)buffer)[i] = template->long_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(long));
				((int *)temp)[0] = template->int_loc_tag_val;
			}
			break;
		case SHARP_DTYPE_UNSIGNED:
			((unsigned int *)buffer)[i] = template->unsigned_int_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				((unsigned int *)buffer)[i] = template->int_loc_tag_val;
			}
			break;
		case SHARP_DTYPE_UNSIGNED_LONG:
			((unsigned long *)buffer)[i] = template->unsigned_long_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(unsigned long));
				((int *)temp)[0] = template->int_loc_tag_val;
			}
			break;
		case SHARP_DTYPE_FLOAT:
			((float *)buffer)[i] = template->float_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				((int *)buffer)[i] = template->int_loc_tag_val;
			}
			break;
		case SHARP_DTYPE_DOUBLE:
			((double *)buffer)[i] = template->double_val;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(double));
				((int *)temp)[0] = template->int_loc_tag_val;
			}
			break;
		default:
			break;
		}
	}
}

/* handle coll request to complete */
static int is_valid_op(enum sharp_datatype type, enum sharp_reduce_op op)
{
	switch(type) {
	case SHARP_DTYPE_INT:
	case SHARP_DTYPE_LONG:
	case SHARP_DTYPE_UNSIGNED:
	case SHARP_DTYPE_UNSIGNED_LONG:
		return 1;
	case SHARP_DTYPE_FLOAT:
	case SHARP_DTYPE_DOUBLE:
		if (op == SHARP_OP_MIN || op == SHARP_OP_MAX ||
		    op == SHARP_OP_SUM || op == SHARP_OP_PROD ||
		    op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC)
			return 1;
	case SHARP_DTYPE_NULL:
	default:
		return 0;
	}

	return 0;
}

static double apply_double_op(enum sharp_reduce_op reduce_op, double x, double y)
{
	double value;
	switch(reduce_op) {
	case SHARP_OP_MAX:
	case SHARP_OP_MAXLOC:
		value = (x > y ? x : y);
		break;
	case SHARP_OP_MIN:
	case SHARP_OP_MINLOC:
		value = (x < y ? x : y);
		break;
	case SHARP_OP_SUM:
		value = x + y;
		break;
	case SHARP_OP_PROD:
		value = x * y;
		break;
	default:
		value = 0.0;
		break;
	}

	return value;
}

static long apply_int_op(enum sharp_reduce_op reduce_op, long x, long y)
{
	long value;
	switch(reduce_op) {
	case SHARP_OP_MAX:
	case SHARP_OP_MAXLOC:
		value = (x > y ? x : y);
		break;
	case SHARP_OP_MIN:
	case SHARP_OP_MINLOC:
		value = (x < y ? x : y);
		break;
	case SHARP_OP_SUM:
		value = x + y;
		break;
	case SHARP_OP_PROD:
		value = x * y;
		break;
	case SHARP_OP_LAND:
		value = x && y;
		break;
	case SHARP_OP_BAND:
		value = x & y;
		break;
	case SHARP_OP_LOR:
		value = x || y;
		break;
	case SHARP_OP_BOR:
		value = x | y;
		break;
	case SHARP_OP_LXOR:
		value = (x || y) && !(x && y);
		break;
	case SHARP_OP_BXOR:
		value = x ^ y;
		break;
	case SHARP_OP_NULL:
	default:
		value = 0;
		break;
	}
	return value;
}

static void test_get_reduce_result(enum sharp_reduce_op reduce_op, int num_procs, struct data_template *template)
{
	long x1 = 1, x2 = 2, iresult_64;
	int iresult_32;
	double dx1 = 1.0, dx2 = 2.0 , dresult = 0.0;

	if (reduce_op == SHARP_OP_MINLOC) {
		template->int_min_max_loc_result = 0;
	} else if (reduce_op == SHARP_OP_MAXLOC) {
		template->int_min_max_loc_result = (num_procs - 1);
	}

	if (num_procs == 1) {
		test_data_template_init(template, 1);
		return;
	}

	iresult_64 = apply_int_op(reduce_op, x1, x2);
	for (x1 = 3, x2 = 3; x2 <= num_procs; x1++, x2++) {
		iresult_64 = apply_int_op(reduce_op, iresult_64, x1);
	}
	iresult_32 = (int )iresult_64;
	template->int_val = iresult_32;
	template->unsigned_int_val = iresult_32;
	template->long_val = iresult_64;
	template->unsigned_long_val = iresult_64;

	if (reduce_op < SHARP_OP_LAND ||
	    reduce_op == SHARP_OP_MINLOC || reduce_op == SHARP_OP_MAXLOC) {
		dresult = apply_double_op(reduce_op, dx1, dx2);
		for (dx2 = 3.0, x2 = 3; x2 <= num_procs; x2++, dx2 += 1.0) {
			dresult = apply_double_op(reduce_op, dresult, dx2);
		}
	}
	template->float_val =  (float) dresult;
	template->double_val =  (double) dresult;
}

static int test_check_buffer_errors(enum sharp_datatype type, int length,
				    struct data_template *template,
				    void *buffer, enum sharp_reduce_op op)
{
	int i, errors = 0;
	void *temp;

	for (i = 0; i < length; i++) {
		switch(type) {
		case SHARP_DTYPE_INT:
			if (((int *)buffer)[i] != template->int_val)
				errors++;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC)
				if (((int *)buffer)[++i] != template->int_min_max_loc_result)
					errors++;

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%d, expected:%d \n",
						 i, ((int *)buffer)[i], template->int_val);
			break;
		case SHARP_DTYPE_LONG:
			if (((long *)buffer)[i] != template->long_val)
				errors++;

			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(long));
				if (((int *)temp)[0] != template->int_min_max_loc_result)
					errors++;
			}

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%ld, expected:%ld \n",
						 i, ((long *)buffer)[i], template->long_val);
			break;
		case SHARP_DTYPE_UNSIGNED:
			if (((unsigned int *)buffer)[i] != template->unsigned_int_val)
				errors++;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC)
				if (((unsigned int *)buffer)[++i] != template->int_min_max_loc_result)
					errors++;

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%d, expected:%d \n",
						 i, ((unsigned int *)buffer)[i], template->unsigned_int_val);
					break;
		case SHARP_DTYPE_UNSIGNED_LONG:
			if (((unsigned long *)buffer)[i] != template->unsigned_long_val)
				errors++;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(long));
				if (((int *)temp)[0] != template->int_min_max_loc_result)
					errors++;
			}

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%lu, expected:%lu \n",
						 i, ((unsigned long *)buffer)[i], template->unsigned_long_val);
			break;
		case SHARP_DTYPE_FLOAT:
			if (((float *)buffer)[i] != template->float_val)
				errors++;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC)
				if (((int *)buffer)[++i] != template->int_min_max_loc_result)
					errors++;

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%10.10f, expected:%10.10f \n",
						 i, ((float *)buffer)[i], template->float_val);
			break;
		case SHARP_DTYPE_DOUBLE:
			if (((double *)buffer)[i] != template->double_val)
				errors++;
			if (op == SHARP_OP_MINLOC || op == SHARP_OP_MAXLOC) {
				i++;
				temp = (buffer + i * sizeof(double));
				if (((int *)temp)[0] != template->int_min_max_loc_result)
					errors++;
			}

			if (errors == 1)
				fprintf(stdout,"Data validation error. pos:%d, value:%10.10f, expected:%10.10f\n",
						 i, ((double *)buffer)[i], template->float_val);
			break;
		case SHARP_DTYPE_NULL:
		default:
			break;
		}
	}
	return errors;
}

void coll_test_sharp_barrier_complex(coll_sharp_module_t *sharp_comm)
{
	int i, size, rank, ret;
	MPI_Comm mpi_comm;

	mpi_comm = sharp_comm->comm;
	MPI_Comm_size(mpi_comm, &size);
	MPI_Comm_rank(mpi_comm, &rank);

	if (rank == 0)
		printf("\nBarrier validation test. expecting rank print in order\n");

	for (i=0; i<size; i++){
		if (rank == i){
			printf("Rank %d\n",rank);
			fflush(stdout);
			usleep(100000);
		}

		if (enable_sharp_coll) {
			ret = sharp_coll_do_barrier(sharp_comm->sharp_coll_comm);
			if (ret != SHARP_COLL_SUCCESS) {
				if (rank == 0)
					fprintf(stderr, "Barrier failed: %s\n",
						sharp_coll_strerror(ret));
			}
		} else {
			MPI_Barrier(mpi_comm);
		}
	}
}

void coll_test_sharp_allreduce_complex(coll_sharp_module_t *sharp_comm)
{
	int *inbuf = NULL, *outbuf = NULL;
	int max_count, count;
	int type, op_count;
	int comm_size, my_rank;
	int errors, ret;
	struct data_template template;
	MPI_Comm mpi_comm =  sharp_comm->comm;
	struct sharp_coll_reduce_spec reduce_spec;

	MPI_Comm_size(mpi_comm, &comm_size);
	MPI_Comm_rank(mpi_comm, &my_rank);

	if (my_rank == 0)
		printf("\nAllreduce validation test with all datatypes and reduce ops.\n");

	inbuf = (int *)malloc(max_payload_size);
	if (!inbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	outbuf = (int *)malloc(max_payload_size);
	if (!inbuf || !outbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	/*loop over all supported datatypes */
	for (type = 0; sharp_test_data_types[type].id != SHARP_DTYPE_NULL; type++) {
		max_count = max_payload_size / sharp_test_data_types[type].size;

		/*loop over all sizes */
		for (count = 1;  count <= max_count; count = (count == 0)? 1 : count * 2) {
			/* loop over all ops */
			for (op_count = 0; sharp_test_ops[op_count].id != SHARP_OP_NULL; op_count++) {

				/* Not all ops are valid on all datatypes*/
				if (!is_valid_op(sharp_test_data_types[type].id, sharp_test_ops[op_count].id))
					continue;

				if (sharp_test_ops[op_count].id == SHARP_OP_MINLOC ||
				    sharp_test_ops[op_count].id == SHARP_OP_MAXLOC) {
					/*min_loc and max_loc need even count */
					if (count % 2)
						continue;
				}
				if (my_rank == 0) {
					fprintf(stdout, "Allreduce Test: data type :%s Op:%s count:%d\n",
						sharp_test_data_types[type].name,
						sharp_test_ops[op_count].name, count);
				}

				/*Initialize input buffers*/
				test_data_template_init(&template, (my_rank + 1));
				test_init_buffer(sharp_test_data_types[type].id, count,
						 &template, inbuf, sharp_test_ops[op_count].id);
				test_data_template_init(&template, -1);
				test_init_buffer(sharp_test_data_types[type].id, count,
						 &template, outbuf, sharp_test_ops[op_count].id);

				/* Do Allreduce */
				if (enable_sharp_coll) {
					reduce_spec.sbuf_desc.buffer.ptr = inbuf;
					reduce_spec.sbuf_desc.buffer.length = (count * sharp_test_data_types[type].size);
					reduce_spec.sbuf_desc.buffer.mem_handle = NULL;
					reduce_spec.sbuf_desc.type = SHARP_DATA_BUFFER;
					reduce_spec.rbuf_desc.buffer.ptr = outbuf;
					reduce_spec.rbuf_desc.buffer.length = (count * sharp_test_data_types[type].size);
					reduce_spec.rbuf_desc.buffer.mem_handle = NULL;
					reduce_spec.rbuf_desc.type = SHARP_DATA_BUFFER;
					reduce_spec.dtype = sharp_test_data_types[type].id;
					reduce_spec.length = count;
					reduce_spec.op = sharp_test_ops[op_count].id;

					ret = sharp_coll_do_allreduce(sharp_comm->sharp_coll_comm, &reduce_spec);
					if (ret != SHARP_COLL_SUCCESS) {
						if (my_rank == 0)
							fprintf(stderr, "Allreduce failed: %s\n",
								sharp_coll_strerror(ret));
					}
				} else {
					MPI_Datatype dt = sharp_test_data_types[type].mpi_id;
					int mpi_count = count;
					if (sharp_test_ops[op_count].id == SHARP_OP_MINLOC ||
					    sharp_test_ops[op_count].id == SHARP_OP_MAXLOC) {
						/*get mic_loc max_loc datatype */
						dt = sharp_test_data_types[type].mpi_min_max_id;
						mpi_count = count / 2;
					}

					MPI_Allreduce(inbuf, outbuf, mpi_count, dt,
						      sharp_test_ops[op_count].mpi_op, mpi_comm);

				}

				/*Validate result buffer*/
				errors = 0;
				test_get_reduce_result(sharp_test_ops[op_count].id, comm_size, &template);
				errors = test_check_buffer_errors(sharp_test_data_types[type].id,
						count, &template, outbuf, sharp_test_ops[op_count].id);
				if (errors) {
					fprintf(stdout,"Allreduce failed. comm:%p count:%d date type:%s reduce Op:%s\n",
						sharp_comm->comm, count, sharp_test_data_types[type].name,
						sharp_test_ops[op_count].name);
				}
			}
		}
	}


fn_fail:
	free(inbuf);
	free(outbuf);

}
void coll_test_sharp_iallreduce_complex(coll_sharp_module_t *sharp_comm)
{
	fprintf(stdout, "iallreduce complex not implemented\n");
}

void coll_test_sharp_barrier_basic(coll_sharp_module_t *sharp_comm)
{
	int rank, ret;
	MPI_Comm mpi_comm =  sharp_comm->comm;
	MPI_Comm_rank(mpi_comm, &rank);

	if (rank == 0)
		printf("Barrier Basic test.\n");

	if (enable_sharp_coll) {
		ret = sharp_coll_do_barrier(sharp_comm->sharp_coll_comm);
		if (ret != SHARP_COLL_SUCCESS) {
			if (rank == 0)
				fprintf(stderr, "Barrier failed: %s\n",
					sharp_coll_strerror(ret));
		}
	} else {
		MPI_Barrier(mpi_comm);
	}
}

void coll_test_sharp_allreduce_basic(coll_sharp_module_t *sharp_comm)
{

	int *inbuf = NULL, *outbuf = NULL;
	int count, result, i, size, rank, ret;
	struct sharp_coll_reduce_spec reduce_spec;
	MPI_Comm mpi_comm;
	void *mem_mr = NULL;

	mpi_comm =  sharp_comm->comm;
	count = max_payload_size / sizeof(int);

	MPI_Comm_size(mpi_comm, &size);
	MPI_Comm_rank(mpi_comm, &rank);

	if (rank == 0)
		printf("Allreduce Basic test (count:%d op:%s).\n", count, "SUM");

	inbuf = (int *)malloc(max_payload_size);
	if (!inbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	outbuf = (int *)malloc (max_payload_size);
	if (!inbuf || !outbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	if (enable_sharp_coll) {
		sharp_coll_reg_mr(coll_sharp_component.sharp_coll_context,
				  inbuf, max_payload_size, &mem_mr);
	}

	for (i = 0; i < count; i++)
		inbuf[i] = (rank + i);

	if (enable_sharp_coll) {
		reduce_spec.sbuf_desc.buffer.ptr = inbuf;
		reduce_spec.sbuf_desc.buffer.length = (count * sizeof(int));
		reduce_spec.sbuf_desc.buffer.mem_handle = mem_mr;
		reduce_spec.sbuf_desc.type = SHARP_DATA_BUFFER;
		reduce_spec.rbuf_desc.buffer.ptr = outbuf;
		reduce_spec.rbuf_desc.buffer.length = (count * sizeof(int));
		reduce_spec.rbuf_desc.buffer.mem_handle = NULL;
		reduce_spec.rbuf_desc.type = SHARP_DATA_BUFFER;
		reduce_spec.dtype = SHARP_DTYPE_INT;
		reduce_spec.length = count;
		reduce_spec.op = SHARP_OP_SUM;

		ret = sharp_coll_do_allreduce(sharp_comm->sharp_coll_comm, &reduce_spec);
		if (ret != SHARP_COLL_SUCCESS) {
			if (rank == 0)
				fprintf(stderr, "Allreduce failed: %s\n",
					sharp_coll_strerror(ret));
		}
	} else {
		MPI_Allreduce(inbuf, outbuf, count, MPI_INT, MPI_SUM, mpi_comm);
	}

	for (i = 0 ; i < count; i++) {
		result = i * size + (size * (size - 1)) / 2;
		if (outbuf[i] !=  result) {
			fprintf(stdout,"data validation failed. At pos:%d got  =%d, expected:%d \n",
				    i, outbuf[i], result);
			break;
		}
	}

fn_fail:
	if (enable_sharp_coll && mem_mr)
		sharp_coll_dereg_mr(coll_sharp_component.sharp_coll_context, mem_mr);
	free(inbuf);
	free(outbuf);
}

void coll_test_sharp_iallreduce_basic(coll_sharp_module_t *sharp_comm)
{

	int *inbuf = NULL, *outbuf = NULL;
	int count, result, i, j, size, rank, ret;
	struct sharp_coll_reduce_spec reduce_spec;
	MPI_Comm mpi_comm;
	MPI_Request mpi_reqs[MAX_NB_REQS];
	MPI_Status mpi_status;
	void *sharp_reqs[MAX_NB_REQS];
	void *mem_mr = NULL;
	int outstanding_non_blocking_ops = nbc_count;

	mpi_comm =  sharp_comm->comm;
	count = max_payload_size / sizeof(int);

	MPI_Comm_size(mpi_comm, &size);
	MPI_Comm_rank(mpi_comm, &rank);

	if (rank == 0)
		printf("IAllreduce Basic test(count:%d op:%s NB_count:%d)\n",
			count, "SUM", nbc_count);

	inbuf = (int *)malloc(max_payload_size * outstanding_non_blocking_ops);
	if (!inbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	outbuf = (int *)malloc (max_payload_size * outstanding_non_blocking_ops);
	if (!inbuf || !outbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	if (enable_sharp_coll) {
		sharp_coll_reg_mr(coll_sharp_component.sharp_coll_context, inbuf,
			outstanding_non_blocking_ops * max_payload_size, &mem_mr);
	}

	for (i = 0; i < outstanding_non_blocking_ops; i++) {
		for (j = 0; j < count; j++) {
			inbuf[(i * count) + j] = (rank + j);
		}
	}

	for (i = 0; i < outstanding_non_blocking_ops; i++) {
		if (enable_sharp_coll) {
			reduce_spec.sbuf_desc.buffer.ptr = inbuf + (i * count);
			reduce_spec.sbuf_desc.buffer.length = (count * sizeof(int));
			reduce_spec.sbuf_desc.buffer.mem_handle = mem_mr;
			reduce_spec.sbuf_desc.type = SHARP_DATA_BUFFER;
			reduce_spec.rbuf_desc.buffer.ptr = outbuf + (i * count);
			reduce_spec.rbuf_desc.buffer.length = (count * sizeof(int));
			reduce_spec.rbuf_desc.buffer.mem_handle = NULL;
			reduce_spec.rbuf_desc.type = SHARP_DATA_BUFFER;
			reduce_spec.dtype = SHARP_DTYPE_INT;
			reduce_spec.length = count;
			reduce_spec.op = SHARP_OP_SUM;

			ret = sharp_coll_do_allreduce_nb(sharp_comm->sharp_coll_comm, &reduce_spec, &sharp_reqs[i]);
			if (ret != SHARP_COLL_SUCCESS) {
				if (rank == 0)
					fprintf(stderr, "Allreduce failed: %s\n",
							sharp_coll_strerror(ret));
			}
		} else {
			MPI_Iallreduce(inbuf + (i * count), outbuf + (i * count), count,
				MPI_INT, MPI_SUM, mpi_comm, &mpi_reqs[i]);
		}
	}

	for (j = 0; j < outstanding_non_blocking_ops; j++) {
		if (enable_sharp_coll) {
			ret = sharp_coll_req_wait(sharp_reqs[j]);
			if (ret != SHARP_COLL_SUCCESS) {
				fprintf(stderr, "sharp_coll_req_wait failed \n");
			}
			sharp_coll_req_free(sharp_reqs[j]);
		} else {
			MPI_Wait(&mpi_reqs[j], &mpi_status);
		}

		for (i = 0 ; i < count; i++) {
			result = i * size + (size * (size - 1)) / 2;
			if (outbuf[(j * count) + i] !=  result) {
				fprintf(stdout,"data validation failed. #op:%d At pos:%d got  =%d, expected:%d\n",
						j, i, outbuf[(j * count) + i], result);
				break;
			}
		}
	}

fn_fail:
	if (enable_sharp_coll && mem_mr)
		sharp_coll_dereg_mr(coll_sharp_component.sharp_coll_context, mem_mr);
	free(inbuf);
	free(outbuf);
}
static void __attribute__((unused)) print_coll_iterations_perf_data(double *all_iter_time, int rank, int comm_size,
				       int data_size, int iterations)
{
	int i, j, k, max_cutoff ;
	double avg_time = 0.0, max_time = 0.0, min_time = 0.0, timer = 0.0;
	double *sum_time = NULL, max_iter_time, min_iter_time, latency;
	char *max_cutoffs, *endptr = NULL, *saveptr = NULL, *str = NULL;

	if (getenv("MAX_CUTOFF_LIST"))
		max_cutoffs= strdup(getenv("MAX_CUTOFF_LIST"));
	else
		max_cutoffs = strdup("10");

	if (!max_cutoffs)
		goto fn_fail;

	sum_time = (double *) malloc(sizeof(double) * comm_size);
	if (!sum_time)
		goto fn_fail;

	saveptr = str = strdup(max_cutoffs);
	for (endptr = strtok_r (str,",", &saveptr); endptr; endptr = strtok_r (NULL, ",", &saveptr)) {
		max_cutoff = atoi(endptr);

		for (j = 0; j < comm_size; j++)
			sum_time[j] = 0.0;
		k = 0;
		max_time = 0.0;
		min_time = 10000.0;
		for (i = 0; i < iterations; i++) {
			max_iter_time = 0.0;
			min_iter_time = 10000.0;
			for (j = 0; j < comm_size; j++) {
				latency = (double)(all_iter_time[i + (j*iterations)] * 1e6);
				if (latency > max_iter_time)
					max_iter_time = latency;
				if (latency < min_iter_time)
					min_iter_time = latency;
			}
			if (max_iter_time > max_cutoff) continue;
			if (max_iter_time > max_time) max_time = max_iter_time;
			if (min_iter_time < min_time) min_time = min_iter_time;

			for (j = 0; j < comm_size; j++) {
				sum_time[j] += (double)(all_iter_time[i + (j*iterations)]);
			}
			k++;
		}
		if (k == 0)
			continue;

		timer=0.0;
		for (j = 0; j < comm_size; j++)
			timer += sum_time[j];
		avg_time =  (double)(timer * 1e6) / (comm_size * k);
		if (data_size < 0)
			fprintf(stdout, "%10.2f %10.2f  %10.2f %10d  %10d\n", avg_time, min_time, max_time, k, max_cutoff);
		else
			fprintf(stdout, "%15d %10.2f %10.2f  %10.2f %10d  %10d\n", data_size,
				avg_time, min_time, max_time, k, max_cutoff);
	}
#if 0
	fprintf(stdout, "Rank  ");
	for (j = 0; j < comm_size; j++)
		fprintf(stdout, "%8d", j);
	for (i = 0; i < iterations; i++) {
		fprintf(stdout, "\nIter:%d", i);
		for (j = 0; j < comm_size; j++) {
			fprintf(stdout, "%8.2f",
				(double)(all_iter_time[i + (j*iterations)] * 1e6));
		}
	}
#endif
	fprintf(stdout, "\n");
fn_fail:
	free(str);
	free(max_cutoffs);
	free(sum_time);
}

void coll_test_sharp_barrier_perf(coll_sharp_module_t *sharp_comm)
{
	int i, j, comm_size, rank, ret;
	double latency = 0.0, t_start = 0.0, t_stop = 0.0;
	double avg_time = 0.0, max_time = 0.0, min_time = 0.0, timer = 0.0;
	double *iter_time, max_iter_time, min_iter_time;
	int iterations = perf_test_iterations, skips = perf_test_skips;
	double *all_iter_time = NULL;
	MPI_Comm mpi_comm;

	mpi_comm =  sharp_comm->comm;
	MPI_Comm_size(mpi_comm, &comm_size);
	MPI_Comm_rank(mpi_comm, &rank);

	if (rank == 0) {
		fprintf(stdout, "\nBarrier perf test. comm_size:%d\n", comm_size);
		fprintf(stdout, "%10s %10s %10s %10s\n", "Avg", "Min", "Max", "iters");
	}

	iter_time = (double *) malloc(sizeof(double) * iterations);
	if (!iter_time) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	if (rank == 0) {
		all_iter_time = (double *) malloc(sizeof(double) * iterations * comm_size);
		if (!all_iter_time) {
			fprintf(stdout,"Failed to allocate \n");
			goto fn_fail;
		}

	}

	timer=0.0;
	for(i=0, j= 0; i < iterations + skips; i++) {

		t_start = sharp_time_sec();

		if (enable_sharp_coll) {
			ret = sharp_coll_do_barrier(sharp_comm->sharp_coll_comm);
			if (ret != SHARP_COLL_SUCCESS) {
				if (rank == 0)
					fprintf(stderr, "Barrier failed: %s\n",
						sharp_coll_strerror(ret));
			}
		} else {
			MPI_Barrier(mpi_comm);
		}
		if (i >= skips) {
			t_stop = sharp_time_sec();
			timer += (t_stop - t_start);
			iter_time[j++] = (t_stop - t_start);
		}
	}

	latency = (double)(timer * 1e6) / iterations;

	min_iter_time = max_iter_time = iter_time[0];
	for (i = 0; i < iterations; i++) {
		if (min_iter_time > iter_time[i])
			min_iter_time = iter_time[i];
		if (max_iter_time < iter_time[i])
			max_iter_time = iter_time[i];
	}
	min_iter_time = (double)(min_iter_time * 1e6);
	max_iter_time = (double)(max_iter_time * 1e6);

	MPI_Reduce(&min_iter_time, &min_time, 1, MPI_DOUBLE, MPI_MIN, 0, mpi_comm);
	MPI_Reduce(&max_iter_time, &max_time, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_comm);
	MPI_Reduce(&latency, &avg_time, 1, MPI_DOUBLE, MPI_SUM, 0, mpi_comm);

	avg_time = avg_time/ comm_size;

	MPI_Gather(iter_time, iterations, MPI_DOUBLE,
		   all_iter_time, iterations, MPI_DOUBLE, 0, mpi_comm);

	if (rank == 0) {
		fprintf(stdout, "%10.2f %10.2f %10.2f %10d\n", avg_time, min_time, max_time, iterations);
#if 0
		print_coll_iterations_perf_data(all_iter_time, rank, comm_size, -1, iterations);
#endif
		fflush(stdout);
	}
fn_fail:
	free(iter_time);
	free(all_iter_time);

}

void coll_test_sharp_allreduce_perf(coll_sharp_module_t *sharp_comm)
{

	float *inbuf = NULL, *outbuf = NULL;
	int count, i, j, comm_size, rank, max_count, ret;
	double latency = 0.0, t_start = 0.0, t_stop = 0.0;
	double avg_time = 0.0, max_time = 0.0, min_time = 0.0, timer = 0.0;
	double *iter_time = NULL, max_iter_time, min_iter_time;
	double *all_iter_time = NULL;
	int iterations = perf_test_iterations, skips = perf_test_skips;
	MPI_Comm mpi_comm;
	struct sharp_coll_reduce_spec reduce_spec;
	void *mem_mr = NULL;

	mpi_comm =  sharp_comm->comm;

	if (mpi_comm != MPI_COMM_WORLD)
		return;

	mallopt(M_TRIM_THRESHOLD, -1);
	mallopt(M_MMAP_MAX, 0);

	max_count = max_payload_size / sizeof(float);

	MPI_Comm_size(mpi_comm, &comm_size);
	MPI_Comm_rank(mpi_comm, &rank);

	inbuf = memalign(getpagesize(), max_payload_size);
	if (!inbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}

	outbuf = memalign(getpagesize(), max_payload_size);
	if (!outbuf) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}


	if (rank == 0) {
		fprintf(stdout, "\nAllreduce perf test. comm_size:%d\n", comm_size);
		fprintf(stdout, "%15s %10s %10s %10s %10s\n", "#size(bytes)", "Avg", "Min", "Max", "iters");
	}

	iter_time = (double *) calloc(1, sizeof(double) * iterations);
	if (!iter_time) {
		fprintf(stdout,"Failed to allocate \n");
		goto fn_fail;
	}
	if (rank == 0) {
		all_iter_time = (double *) calloc(1, sizeof(double) * iterations * comm_size);
		if (!all_iter_time) {
			fprintf(stdout,"Failed to allocate \n");
			goto fn_fail;
		}
	}

	if (mlockall(MCL_FUTURE|MCL_CURRENT) < 0) {
		fprintf(stdout,"mlockall failed: %m\n");
	}

	if (enable_sharp_coll) {
		sharp_coll_reg_mr(coll_sharp_component.sharp_coll_context,
					inbuf, max_payload_size, &mem_mr);
	}

	for (count = 1; count <= max_count; count *= 2) {

		MPI_Barrier(mpi_comm);
		timer = 0.0;
		for (i = 0, j = 0; i < iterations + skips; i++) {
#if _DCHECK
			int pos, result;
			for (pos = 0; pos < count; pos++) {
				inbuf[pos] = i;
				outbuf[pos] = 0;
			}
#endif
			if (perf_with_barrier) {
				if (enable_sharp_coll) {
					ret = sharp_coll_do_barrier(sharp_comm->sharp_coll_comm);
					if (ret != SHARP_COLL_SUCCESS) {
						if (rank == 0)
							fprintf(stderr, "Barrier failed: %s\n",
								sharp_coll_strerror(ret));
					}
				} else {
					MPI_Barrier(mpi_comm);
				}
			}

			t_start = sharp_time_sec();

			if (enable_sharp_coll) {
				reduce_spec.sbuf_desc.buffer.ptr = inbuf;
				reduce_spec.sbuf_desc.buffer.length = (count * sizeof(int));
				reduce_spec.sbuf_desc.buffer.mem_handle = mem_mr;
				reduce_spec.sbuf_desc.type = SHARP_DATA_BUFFER;
				reduce_spec.rbuf_desc.buffer.ptr = outbuf;
				reduce_spec.rbuf_desc.buffer.length = (count * sizeof(int));
				reduce_spec.rbuf_desc.buffer.mem_handle = NULL;
				reduce_spec.rbuf_desc.type = SHARP_DATA_BUFFER;
				reduce_spec.dtype = SHARP_DTYPE_FLOAT;
				reduce_spec.length = count;
				reduce_spec.op = SHARP_OP_SUM;

				ret = sharp_coll_do_allreduce(sharp_comm->sharp_coll_comm, &reduce_spec);
				if (ret != SHARP_COLL_SUCCESS) {
					if (rank == 0)
						fprintf(stderr, "Allreduce failed: %s\n",
							sharp_coll_strerror(ret));
				}
			} else {
				MPI_Allreduce(inbuf, outbuf, count, MPI_FLOAT, MPI_SUM, mpi_comm);
			}

			if (i >= skips) {
				t_stop = sharp_time_sec();
				timer += (t_stop - t_start);
				iter_time[j++] = (t_stop - t_start);
			}
#if _DCHECK
			for (pos = 0 ; pos < count; pos++) {
				result = i * comm_size;
				if (outbuf[pos] !=  result) {
					fprintf(stdout,"data validation failed. At pos:%d got  =%d, expected:%d\n",
							 pos, outbuf[pos], result);
					break;
				}
			}
#endif


		}

		latency = (double)(timer * 1e6) / iterations;

		min_iter_time = max_iter_time = iter_time[0];
		for (i = 0; i < iterations; i++) {
			if (min_iter_time > iter_time[i])
				min_iter_time = iter_time[i];
			if (max_iter_time < iter_time[i])
				max_iter_time = iter_time[i];
		}
		min_iter_time = (double)(min_iter_time * 1e6);
		max_iter_time = (double)(max_iter_time * 1e6);

		MPI_Reduce(&min_iter_time, &min_time, 1, MPI_DOUBLE, MPI_MIN, 0, mpi_comm);
		MPI_Reduce(&max_iter_time, &max_time, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_comm);
		MPI_Reduce(&latency, &avg_time, 1, MPI_DOUBLE, MPI_SUM, 0, mpi_comm);

		avg_time = avg_time/ comm_size;

		MPI_Gather(iter_time, iterations, MPI_DOUBLE,
			   all_iter_time, iterations, MPI_DOUBLE, 0, mpi_comm);
		if (rank == 0) {
			fprintf(stdout, "%15d %10.2f %10.2f %10.2f %10d\n", (int )(count * sizeof(int)), avg_time, min_time, max_time, iterations);
#if 0
			print_coll_iterations_perf_data(all_iter_time, rank, comm_size, (int )(count * sizeof(int)), iterations);
#endif
			fflush(stdout);
		}
	}

fn_fail:
	if (enable_sharp_coll && mem_mr) {
		sharp_coll_dereg_mr(coll_sharp_component.sharp_coll_context, mem_mr);
	}
	free(inbuf);
	free(outbuf);
	free(iter_time);
	free(all_iter_time);
}
void coll_test_sharp_iallreduce_perf(coll_sharp_module_t *sharp_comm)
{
	fprintf(stdout, "iallreduce perf is not implemented");
}
